@@ -1,4 +1,4 @@
-# Copyrights 2008-2013 by [Mark Overmeer].
+# Copyrights 2008-2014 by [Mark Overmeer].
 #  For other contributors see ChangeLog.
 # See the manual pages for details on the licensing terms.
 # Pod stripped from pm file by OODoc 2.01.
@@ -7,7 +7,7 @@ use strict;
 
 package XML::Compile::Cache;
 use vars '$VERSION';
-$VERSION = '1.00';
+$VERSION = '1.03';
 
 use base 'XML::Compile::Schema';
 
@@ -30,19 +30,17 @@ sub init($)
     $self->{XCC_wopts}  = delete $args->{opts_writers} || [];
     $self->{XCC_undecl} = delete $args->{allow_undeclared} || 0;
 
-    $self->{XCC_rcode}  = {};  # compiled code refs
-    $self->{XCC_wcode}  = {};
     $self->{XCC_dropts} = {};  # declared opts
     $self->{XCC_dwopts} = {};
     $self->{XCC_uropts} = {};  # undeclared opts
     $self->{XCC_uwopts} = {};
 
-    $self->{XCC_readers} = {};
+    $self->{XCC_readers} = {}; # compiled code refs;
     $self->{XCC_writers} = {};
 
     $self->typemap($args->{typemap});
     $self->xsiType($args->{xsi_type});
-    $self->anyElement($args->{any_element} || 'SKIP_ALL');
+    $self->anyElement($args->{any_element} || 'ATTEMPT');
 
     $self;
 }
@@ -60,7 +58,7 @@ sub typemap(@)
 }
 
 
-sub xsiType(@)
+sub addXsiType(@)
 {   my $self = shift;
     my $x    = $self->{XCC_xsi_type} ||= {};
     my @d    = @_ > 1 ? @_ : !defined $_[0] ? ()
@@ -79,6 +77,7 @@ sub xsiType(@)
 
     $x;
 }
+*xsiType = \&addXsiType;
 
 
 sub allowUndeclared(;$)
@@ -127,7 +126,7 @@ sub addPrefixes(@)
 
         if(my $def = $a->{$prefix})
         {   if($def->{uri} ne $ns)
-            {   error __x"prefix {prefix} already refers to {uri}, cannot use it for {ns}"
+            {   error __x"prefix `{prefix}' already refers to {uri}, cannot use it for {ns}"
                   , prefix => $prefix, uri => $def->{uri}, ns => $ns;
             }
         }
@@ -161,6 +160,26 @@ sub prefixFor($)
 }
 
 
+sub addNicePrefix($$)
+{   my ($self, $base, $ns) = @_;
+    if(my $def = $self->prefix($base))
+    {   return $base if $def->{uri} eq $ns;
+    }
+    else
+    {   $self->addPrefixes($base => $ns);
+        return $base;
+    }
+
+    $base .= '01' if $base !~ m/[0-9]$/;
+    while(my $def = $self->prefix($base))
+    {   return $base if $def->{uri} eq $ns;
+        $base++;
+    }
+    $self->addPrefixes($base => $ns);
+    $base;
+}
+
+
 sub learnPrefixes($)
 {   my ($self, $node) = @_;
     my $namespaces = $self->prefixes;
@@ -200,7 +219,8 @@ sub prefixed($;$)
     $ns or return $local;
     my $prefix = $self->prefixFor($ns);
     defined $prefix
-        or error __x"no prefix known for namespace {ns}", ns => $ns;
+        or error __x"no prefix known for namespace `{ns}', use addPrefixes()"
+            , ns => $ns;
 
     length $prefix ? "$prefix:$local" : $local;
 }
@@ -218,7 +238,7 @@ sub compileAll(;$$)
             {   my ($myns, $local) = unpack_type $type;
                 next if $usens eq $myns;
             }
-            $self->{XCC_rcode}{$type} ||= $self->compile(READER=>$type);
+            $self->reader($type);
         }
     }
 
@@ -228,7 +248,7 @@ sub compileAll(;$$)
             {   my ($myns, $local) = unpack_type $type;
                 next if $usens eq $myns;
             }
-            $self->{XCC_wcode}{$type} ||= $self->compile(WRITER => $type);
+            $self->writer($type);
         }
     }
 }
@@ -246,6 +266,7 @@ sub _same_params($$)
 
 sub reader($@)
 {   my ($self, $name) = (shift, shift);
+    my %args    = @_;
     my $type    = $self->findName($name);
     my $readers = $self->{XCC_readers};
 
@@ -260,7 +281,9 @@ sub reader($@)
     {   if(my $ur = $self->{XCC_uropts}{$type})
         {   # do not use cached version when options differ
             _same_params $ur, \@_
-                or return $self->compile(READER => $type, @_);
+                or return $args{is_type}
+                  ? $self->compileType(READER => $type, @_)
+                  : $self->compile(READER => $type, @_);
         }
         else
         {   $self->{XCC_uropts}{$type} = \@_;
@@ -270,13 +293,16 @@ sub reader($@)
          { error __x"type {name} is only declared as writer", name => $name }
     else { error __x"type {name} is not declared", name => $name }
 
-    $readers->{$type} ||= $self->compile(READER => $type, @_);
+    $readers->{$type} ||= $args{is_type}
+       ? $self->compileType(READER => $type, @_)
+       : $self->compile(READER => $type, @_);
 }
 
 
-sub writer($)
+sub writer($%)
 {   my ($self, $name) = (shift, shift);
-    my $type = $self->findName($name);
+    my %args    = @_;
+    my $type    = $self->findName($name);
     my $writers = $self->{XCC_writers};
 
     if(exists $self->{XCC_dwopts}{$type})
@@ -290,7 +316,9 @@ sub writer($)
     {   if(my $ur = $self->{XCC_uwopts}{$type})
         {   # do not use cached version when options differ
             _same_params $ur, \@_
-                or return $self->compile(WRITER => $type, @_)
+                or return $args{is_type}
+                  ? $self->compileType(WRITER => $type, @_)
+                  : $self->compile(WRITER => $type, @_);
         }
         else
         {   $self->{XCC_uwopts}{$type} = \@_;
@@ -303,7 +331,10 @@ sub writer($)
     {   error __x"type {name} is not declared", name => $name;
     }
 
-    $writers->{$type} ||= $self->compile(WRITER => $type, @_);
+    $writers->{$type} ||= $args{is_type}
+       ? $self->compileType(WRITER => $type, @_)
+       : $self->compile(WRITER => $type, @_);
+
 }
 
 sub template($$@)
@@ -406,11 +437,16 @@ sub _cleanup_hooks($)
 {   my ($self, $hooks) = @_;
     $hooks or return;
 
+    # translate prefixed type names into full names
     foreach my $hook (ref $hooks eq 'ARRAY' ? @$hooks : $hooks)
-    {   my $types = $hook->{type} or next;
-        $hook->{type} =
-           [ map {ref $_ eq 'Regexp' ? $_ : $self->findName($_)}
+    {   if(my $types = $hook->{type})
+        {   $hook->{type} =
+              [ map {ref $_ eq 'Regexp' ? $_ : $self->findName($_)}
                        ref $types eq 'ARRAY' ? @$types : $types ];
+        }
+        elsif(my $ext = $hook->{extends})
+        {   $hook->{extends} = $self->findName($ext);
+        }
     }
     $hooks;
 }
@@ -546,6 +582,9 @@ sub printIndex(@)
 sub _convertAnyTyped(@)
 {   my ($self, $type, $nodes, $path, $read) = @_;
 
+{no warnings;
+defined $read or warn join ';', caller(0);
+}
     my $key     = $read->keyRewrite($type);
     my $reader  = try { $self->reader($type) };
     if($@)
